#define _POSIX_C_SOURCE 199309L
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#define LENGTH 6400000
#define REPS 100
#define DATA_T float
#define PRINTF_DATA "%.10g"
#define RANDOM_EXPR ((float)rand() / (float)RAND_MAX * 2.0 - 1.0)
#define ABS fabs
/*
#define DATA_T int64_t
#define PRINTF_DATA "%ld"
#define RANDOM_EXPR (rand() % 256 - 128)
#define ABS llabs
*/

void test_array_op(char* name, DATA_T(*array_sum)(DATA_T*, uint64_t), DATA_T* arr, uint64_t n) {
    struct timespec start, end;
    DATA_T total = 0;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    for (uint64_t i = 0; i < REPS; i++) {
        total += array_sum(arr, n);
    }
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) * 1e9 + (end.tv_nsec - start.tv_nsec);
    printf("%-30s%5.2f ns/elt   Result: "PRINTF_DATA"\n", name, elapsed / (REPS * n), total / REPS);
}

DATA_T* random_array(uint64_t n) {
    DATA_T* arr = (DATA_T*)malloc(n * sizeof(DATA_T));
    for (uint64_t i = 0; i < n; i++) {
        arr[i] = RANDOM_EXPR;
    }
    return arr;
}

DATA_T array_sum_1(DATA_T* arr, uint64_t n) {
    DATA_T total = 0;
    for (uint64_t i = 0; i < n; i++) {
        total += arr[i];
    }
    return total;
}

DATA_T array_sum_2(DATA_T* arr, uint64_t n) {
    DATA_T total = 0; 
    for (uint64_t i = 0; i < n - 1; i += 2) {
        total += arr[i];
        total += arr[i + 1];
    }
    return total;
}

DATA_T array_sum_3(DATA_T* arr, uint64_t n) {
    DATA_T total1 = 0, total2 = 0;
    for (uint64_t i = 0; i < n - 1; i += 2) {
        total1 += arr[i];
        total2 += arr[i + 1];
    }
    return total1 + total2;
}

DATA_T array_sum_4(DATA_T* arr, uint64_t n) {
    DATA_T total = 0;
    for (uint64_t i = 0; i < n - 1; i += 2) {
        total += arr[i] + arr[i + 1];
    }
    return total;
}

DATA_T sum_negative(DATA_T* arr, uint64_t n) {
    DATA_T total = 0;
    for (uint64_t i = 0; i < n; i++) {
        if ( arr[i] < 0 ) {
            total += arr[i];
        }
    }
    return total;
}

int main(void) {
    DATA_T* arr = random_array(LENGTH);
    test_array_op("array_sum_1", array_sum_1, arr, LENGTH);
    test_array_op("array_sum_2", array_sum_2, arr, LENGTH);
    test_array_op("array_sum_3", array_sum_3, arr, LENGTH);
    test_array_op("array_sum_4", array_sum_4, arr, LENGTH);
    test_array_op("sum_negative", sum_negative, arr, LENGTH);
    // make the array all negative
    for (uint64_t i = 0; i < LENGTH; i++) {
        arr[i] = -ABS(arr[i]);
    }
    test_array_op("sum_negative (all negative)", sum_negative, arr, LENGTH);
    free(arr);

    arr = random_array(LENGTH);
    test_array_op("sum_negative (half random)", sum_negative, arr, LENGTH);
    // make the array positive/negative in a pattern
    for (uint64_t i = 0; i < LENGTH; i++) {
        if ( i % 2 != 0 ) { // every other is positive
            arr[i] = ABS(arr[i]);
        } else {
            arr[i] = -ABS(arr[i]);
        }
    }
    test_array_op("sum_negative (alternating)", sum_negative, arr, LENGTH);
    free(arr);
    return 0;
}

int main_perf(int argc, char* argv[]) {
    // this main is used for the perf demos, to simplify things.
    DATA_T* arr = random_array(LENGTH);
    DATA_T* arr_alternate = (DATA_T*)malloc(LENGTH * sizeof(DATA_T));
    if (arr_alternate == NULL) {
        return 1;
    }
    for (uint64_t i = 0; i < LENGTH; i++) {
        if ( i % 2 != 0 ) { // every other is positive
            arr_alternate[i] = ABS(arr[i]);
        } else {
            arr_alternate[i] = -ABS(arr[i]);
        }
    }
    if ( argc == 1 || strcmp(argv[1], "1") == 0  ) {
        test_array_op("array_sum_2", array_sum_2, arr, LENGTH);
    } else if ( strcmp(argv[1], "2") == 0  ) {
        test_array_op("array_sum_3", array_sum_3, arr, LENGTH);
    } else if ( strcmp(argv[1], "3") == 0  ) {
        test_array_op("sum_negative arr", sum_negative, arr, LENGTH);
    } else if ( strcmp(argv[1], "4") == 0 ) {
        test_array_op("sum_negative arr_alternate", sum_negative, arr_alternate, LENGTH);
    }
    free(arr);
    free(arr_alternate);
    return 0;
}
